@@ -14,6 +14,7 @@
 #include <pango/pango.h>
 
 #include <cddb-slave-client.h>
+#include <gconf/gconf-client.h>
 
 #include "gnome-cd-type.h"
 #include "cd-selection.h"
@@ -112,6 +113,7 @@ void gnome_cd_set_window_title (GnomeCD *gcd,
 void gnome_cd_build_track_list_menu (GnomeCD *gcd);
 
 void gcd_warning (const char *message, GError *error);
+void gcd_error (const char *message, GError *error);
 void gcd_debug (const gchar *format, ...) G_GNUC_PRINTF (1, 2);
 
 /* theme.c */
@@ -121,8 +123,9 @@ void theme_change_widgets (GnomeCD *gcd);
 void theme_free (GCDTheme *theme);
 
 void make_popup_menu (GnomeCD *gcd,
-                GdkEventButton *event);
+                GdkEventButton *event, gboolean iconify);
 
 void tray_icon_create (GnomeCD *gcd);
+GConfClient *client;
 
 #endif
