@@ -47,7 +47,7 @@ static void
 dialog_response (GtkDialog * widget, gint response_id, GladeXML * dialog)
 {
 	if (response_id == GTK_RESPONSE_HELP)
-		;		/* Launch help for the capplet */
+		gnome_help_display ("gstreamer-properties.xml", NULL, NULL);
 	else
 		gtk_main_quit ();
 }
@@ -259,7 +259,7 @@ init_pipeline_editor(GladeXML * dialog, GSTPPipelineEditor* editor)
 	}
 }
 
-void create_dialog ()
+void create_dialog (void)
 {
 	int i = 0;
 	GdkPixbuf* icon = NULL;
@@ -283,7 +283,7 @@ void create_dialog ()
 	}
 	else {
 		/* FIXME:warning */
-		g_print("Error loading main window icon %s", GSTPROPS_ICONDIR "/gstreamer-properties.png");
+		g_print("Error loading main window icon %s", GSTPROPS_ICONDIR "/gstreamer-properties.png\n");
 	}
 	gtk_widget_show (GTK_WIDGET (main_window));
 }
@@ -296,7 +296,7 @@ main (int argc, char **argv)
 	textdomain (GETTEXT_PACKAGE);
 
 	gnome_program_init ("gstreamer-properties", VERSION, LIBGNOMEUI_MODULE,
-			    argc, argv, NULL);
+			    argc, argv, GNOME_PARAM_APP_DATADIR, DATADIR, NULL);
 	gst_init_with_popt_table (&argc, &argv, NULL);
 
 	/* FIXME: hardcode uninstalled path here */
@@ -310,7 +310,7 @@ main (int argc, char **argv)
 	if (!interface_xml) {
 		/* Fatal error */
 		char *err = g_strdup_printf (_("Could not load UI resource %s"),"gstreamer-properties.glade");
-		g_print ("Error: could not load glade file:\n","gstreamer-properties.glade");
+		g_print ("Error: could not load glade file gstreamer-properties.glade\n");
 		gnome_app_error (GNOME_APP (gnome_program_get ()), err);
 		return 1;
 	}
